/*
 *  Copyright (C) 1997, 1998 Olivetti & Oracle Research Laboratory
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * sockets.c - functions to deal with sockets.
 */

#include <errno.h>
#include <assert.h>
#ifdef USE_WINSOCK
#include "winsock.h"
#include <netinet/in.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#endif
#include <vncview.h>

/*
 * socket buffer routines (NB: global, don't use for different sockets!)
 * The buffer must be large enough to fully accomodate the largest request
 * For non-multithreaded clients, large buffers can help avoid blocking
 * of other clients which share the same display (if the non-mt client is
 * slow at updating the screen).
 */
#define SOCKET_BUFSIZE (512<<10)
static char sbuf[SOCKET_BUFSIZE];
static int sbuf_lo=0, sbuf_hi=0, sbuf_size=0;
static sockType sbuf_sock;
static int use_wattcp=0, use_winsock=1;
#define MIN(_a,_b) ((_a)<(_b)?(_a):(_b))
#define MAX(_a,_b) ((_a)>(_b)?(_a):(_b))

Bool sockets_init(void)
{
#ifdef USE_WINSOCK
        int s;
        if ((s=socket(AF_INET,SOCK_STREAM,0))<0)
        {
                perror("socket");
                return False;
        }
        close(s);
        /* hmm... */
        return True;
#else
        if (sock_init())
        {
                return True;
        }
#endif
        return False;
}

int socket_data_available(sockType sock)
{
        if (sbuf_size > 0) return sbuf_size;
        else
	{
#ifdef USE_WINSOCK
	int w;
	if (fcntl(sock, FIONREAD, &w) < 0) {
	    perror("fcntl");
	    exit(1);
	}

        fprintf(stderr,"%d waiting\n",w);
	return w;

#else
	tcp_tick(sock);
        return sock_dataready(sock);
#endif
	}
}

void read_buffered_exact(sockType sock, char *buf, int n)
{
        int l;

        assert(sbuf_size == ((sbuf_hi-sbuf_lo+SOCKET_BUFSIZE) % SOCKET_BUFSIZE));
        /* first, read enough bytes from the network (blocking) */
        while(sbuf_size<n)
        {
                int can_read, needed, r;
                needed = n-sbuf_size;
                can_read = MIN(SOCKET_BUFSIZE-sbuf_size,SOCKET_BUFSIZE-sbuf_hi);

#ifdef USE_WINSOCK
		l = MIN(can_read,needed);
		r = read(sock,sbuf+sbuf_hi,l);
#else
                l = MIN(can_read,MAX(needed,sock_dataready(sock)));
                r = sock_read(sock, sbuf+sbuf_hi, l);
#endif
#if 0
                fprintf(stderr,"Read %d from network\n",r);
#endif
                if (r < 0)
                {
                        exit(3); /* yuck */
                }
                sbuf_size += r;
                sbuf_hi += r;
                if (sbuf_hi == SOCKET_BUFSIZE)
                        sbuf_hi = 0;
        }
        l = MIN(n,SOCKET_BUFSIZE-sbuf_lo);
        memcpy(buf,sbuf+sbuf_lo,l);
#if 0
        fprintf(stderr,"returning %d from buffer (%d/%d)\n",n,sbuf_lo,sbuf_hi);
#endif
        n -= l;
        sbuf_lo += l;
        sbuf_size -= l;
        if (n>0)
        {
                memcpy(buf+l,sbuf,n);
                sbuf_lo += n;
                sbuf_size -= n;
        }
        sbuf_lo %= SOCKET_BUFSIZE;

        if (sbuf_size == 0) sbuf_lo=sbuf_hi=0;
}

void PrintInHex(char *buf, int len);

Bool errorMessageFromReadExact = True;

/*
 * Read an exact number of bytes, and don't return until you've got them.
 */

Bool
ReadExact(sockType sock, char *buf, int n)
{
    read_buffered_exact(sock,buf,n);
    if (debug)
	PrintInHex(buf,n);
    return True;
}


/*
 * Write an exact number of bytes, and don't return until you've sent them.
 */

Bool
WriteExact(sockType sock, char *buf, int n)
{
    int i = 0;
    int j;

    while (i < n) {
#ifdef USE_WINSOCK
	j = write(sock, buf+i,n-i);
#else
	j = sock_write(sock, buf + i, (n - i));
#endif
	if (j <= 0) {
	    if (j < 0) {
		fprintf(stderr,programName);
		perror(": write");
	    } else {
		fprintf(stderr,"%s: write failed\n",programName);
	    }
	    return False;
	}
	i += j;
    }
    return True;
}


/*
 * ConnectToTcpAddr connects to the given TCP port.
 */

sockType
ConnectToTcpAddr(unsigned int host, int port)
{
	sockType sock;
	int status=0;

#ifdef USE_WINSOCK

    struct sockaddr_in addr;
    int one = 1;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = host;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	fprintf(stderr,programName);
	perror(": ConnectToTcpAddr: socket");
	return -1;
    }

    if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
	fprintf(stderr,programName);
	perror(": ConnectToTcpAddr: connect");
	close(sock);
	return -1;
    }

    if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
		   (char *)&one, sizeof(one)) < 0) {
	fprintf(stderr,programName);
	perror(": ConnectToTcpAddr: setsockopt");
	close(sock);
	return -1;
    }

     
#else
	sock = malloc(sizeof(*sock));
   
        if (!tcp_open(sock, 0, host, port, NULL)) return NULL;
	
	sock_wait_established (sock, sock_delay, NULL, &status);

#endif
	return sock;
	
sock_err:
	return NULL;
}



/*
 * ListenAtTcpPort starts listening at the given TCP port.
 */

int
ListenAtTcpPort(int port)
{
#if 0
    int sock;
    struct sockaddr_in addr;
    int one = 1;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = INADDR_ANY;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	fprintf(stderr,programName);
	perror(": ListenAtTcpPort: socket");
	return -1;
    }

    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
		   (const char *)&one, sizeof(one)) < 0) {
	fprintf(stderr,programName);
	perror(": ListenAtTcpPort: setsockopt");
	close(sock);
	return -1;
    }

    if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
	fprintf(stderr,programName);
	perror(": ListenAtTcpPort: bind");
	close(sock);
	return -1;
    }

    if (listen(sock, 5) < 0) {
	fprintf(stderr,programName);
	perror(": ListenAtTcpPort: listen");
	close(sock);
	return -1;
    }

    return sock;
#endif
}



/*
 * AcceptTcpConnection accepts a TCP connection.
 */

int
AcceptTcpConnection(int listenSock)
{
#if 0
	   int sock;
    struct sockaddr_in addr;
    int addrlen = sizeof(addr);
    int one = 1;

    sock = accept(listenSock, (struct sockaddr *) &addr, &addrlen);
    if (sock < 0) {
	fprintf(stderr,programName);
	perror(": AcceptTcpConnection: accept");
	return -1;
    }

    if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
		   (char *)&one, sizeof(one)) < 0) {
	fprintf(stderr,programName);
	perror(": AcceptTcpConnection: setsockopt");
	close(sock);
	return -1;
    }

    return sock;
#endif
}


/*
 * StringToIPAddr - convert a host string to an IP address.
 */

int
StringToIPAddr(const char *str, unsigned int *addr)
{
#ifdef USE_WINSOCK
    *addr = gethostbyname(str);
#else
    *addr = resolve(str);
#endif

    return (*addr != 0);
}


/*
 * Test if the other end of a socket is on the same machine.
 */

Bool
SameMachine(sockType sock)
{
#if 0
    struct sockaddr_in peeraddr, myaddr;
    int addrlen = sizeof(struct sockaddr_in);

    getpeername(sock, (struct sockaddr *)&peeraddr, &addrlen);
    getsockname(sock, (struct sockaddr *)&myaddr, &addrlen);

    return (peeraddr.sin_addr.s_addr == myaddr.sin_addr.s_addr);
#endif
    return False;
}


/*
 * Print out the contents of a packet for debugging.
 */

void
PrintInHex(char *buf, int len)
{
    int i, j;
    char c, str[17];

    str[16] = 0;

    fprintf(stderr,"ReadExact: ");

    for (i = 0; i < len; i++)
    {
	if ((i % 16 == 0) && (i != 0)) {
	    fprintf(stderr,"           ");
	}
	c = buf[i];
	str[i % 16] = (((c > 31) && (c < 127)) ? c : '.');
	fprintf(stderr,"%02x ",(unsigned char)c);
	if ((i % 4) == 3)
	    fprintf(stderr," ");
	if ((i % 16) == 15)
	{
	    fprintf(stderr,"%s\n",str);
	}
    }
    if ((i % 16) != 0)
    {
	for (j = i % 16; j < 16; j++)
	{
	    fprintf(stderr,"   ");
	    if ((j % 4) == 3) fprintf(stderr," ");
	}
	str[i % 16] = 0;
	fprintf(stderr,"%s\n",str);
    }

    fflush(stderr);
}
